import torch

from .tasks import (
    KoNTATranslationFactory,
    KoNTAAugmentationFactory
)

from .const import LANG_ALIASES


# Task list
SUPPORTED_TASKS = {
    "mt": KoNTATranslationFactory,
    "translation": KoNTATranslationFactory,
    "aug": KoNTAAugmentationFactory,
    "augumentation": KoNTAAugmentationFactory
}


class KoNTA:
    def __new__(
            cls,
            task: str,
            src: str
            ):
        
        if task not in SUPPORTED_TASKS:
            raise KeyError("Unknown task {}, available tasks are {}".format(
                task,
                list(SUPPORTED_TASKS.keys()),
            ))

        if src not in LANG_ALIASES:
            raise KeyError("Unknown target language {}, available target languages are {}".format(
                task,
                list(LANG_ALIASES.keys()),
            ))
        
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        task_module = SUPPORTED_TASKS[task](
            task,
            LANG_ALIASES[src],
            LANG_ALIASES
        ).load(device)

        return task_module

    @staticmethod
    def available_tasks() -> str:
        """
        Returns available tasks in KoNTA project

        Returns:
            str: Supported task names

        """
        return "Available tasks are {}".format(list(SUPPORTED_TASKS.keys()))
    
    @staticmethod
    def available_lang() -> str:
        """
        Returns available language in KoNTA project

        Returns:
            str: Supported language names

        """
        return "Available tasks are {}".format(list(LANG_ALIASES.keys()))


from transformers import NllbTokenizer, AutoModelForSeq2SeqLM
import torch
from transformers import BatchEncoding

class KoNTATranslationFactory():
    """
    Machine translation using NLLN Meta model

    - dataset: Train

    Args:
        src (str): source language
        
    Returns:
        str: machine translation Class

    Examples:
        >>> mt = KoNTA(task="translation", tgt="en")
    """
    def __init__(
            self,
            task: str,
            src: str,
            LANG_ALIASES: dict
            ):
        super().__init__()
        self.task = task
        self.src = src
        self.LANG_ALIASES = LANG_ALIASES

    def load(self, device: str):
        if self.src == "eng_Latn":
            tokenizer = NllbTokenizer.from_pretrained("KoJLabs/nllb-finetuned-en2ko")
            model = AutoModelForSeq2SeqLM.from_pretrained("KoJLabs/nllb-finetuned-en2ko").to(device)
            
        if self.src == "kor_Hang":
            tokenizer = NllbTokenizer.from_pretrained("KoJLabs/nllb-finetuned-ko2en", src_lang=self.src)
            model = AutoModelForSeq2SeqLM.from_pretrained("KoJLabs/nllb-finetuned-ko2en").to(device)

        return KoNTATranslation(
            model,
            tokenizer,
            device,
            self.LANG_ALIASES
        )
    

class KoNTATranslation:
    def __init__(self, model, tokenizer, device, LANG_ALIASES) -> None:
        self.model = model
        self.tokenizer = tokenizer
        self.device = device
        self.LANG_ALIASES = LANG_ALIASES

    def predict(self, text, tgt):
        """
        Predict a translation result

        Args:
            text (str): input text
            tgt (str): target language

        Returns:
            output (str): A translation result
        """
        
        inputs = self.tokenizer(text, padding=True, truncation=True, return_tensors="pt")

        if isinstance(inputs['input_ids'], list):
            input_dict = {}            
            input_dict['input_ids'] = torch.LongTensor(inputs['input_ids']).reshape(1,len(inputs['input_ids']))
            input_dict['attention_mask'] = torch.LongTensor(inputs['attention_mask']).reshape(1,len(inputs['attention_mask']))
            inputs = BatchEncoding(input_dict)
            
        translated_tokens = self.model.generate(
            **inputs.to(self.device), forced_bos_token_id=self.tokenizer.lang_code_to_id[self.LANG_ALIASES[tgt]], max_length=128
        )

        output = self.tokenizer.batch_decode(translated_tokens, skip_special_tokens=True)
            
        return output
    




from transformers import NllbTokenizer, AutoModelForSeq2SeqLM

class KoNTAAugmentationFactory():

    def __init__(self, task, src, LANG_ALIASES: dict):
        super().__init__()
        self.task = task
        self.src = src
        self.LANG_ALIASES = LANG_ALIASES

    def load(self, device):
        tokenizer = NllbTokenizer.from_pretrained("KoJLabs/nllb-finetuned-ko2en")

        ko2en_model = AutoModelForSeq2SeqLM.from_pretrained("KoJLabs/nllb-finetuned-ko2en").to(device)
        en2ko_model = AutoModelForSeq2SeqLM.from_pretrained("KoJLabs/nllb-finetuned-en2ko").to(device)

        return KoNTAAugmentation(
            tokenizer,
            ko2en_model, en2ko_model,
            device,
            self.LANG_ALIASES
        )


class KoNTAAugmentation:

    def __init__(self, 
                 tokenizer,
                 ko2en_model, en2ko_model, 
                 device, LANG_ALIASES):
        self.tokenizer = tokenizer

        self.ko2en_model = ko2en_model
        self.en2ko_model = en2ko_model

        self.device = device
        self.LANG_ALIASES = LANG_ALIASES

    def predict(self, text):

        # ko2en
        translation = self._translate(text, 'eng_Latn', self.tokenizer, self.ko2en_model)

        # en2ko
        backtranslation = self._translate(translation, 'kor_Hang', self.tokenizer, self.en2ko_model)

        return backtranslation


    def _translate(self, text, tgt, tokenizer, model):

        # translation
        inputs = tokenizer(text, padding=True, truncation=True, return_tensors="pt")
        inputs = inputs.to(self.device)
        translated_tokens = model.generate(
            inputs['input_ids'], forced_bos_token_id=tokenizer.lang_code_to_id[tgt], max_length=128
        )

        translation = tokenizer.batch_decode(translated_tokens, skip_special_tokens=True)
        
        return translation


        